﻿using HEAL.Attic;
using HeuristicLab.Algorithms.GeneticAlgorithm;
using HeuristicLab.Algorithms.OffspringSelectionGeneticAlgorithm;
using HeuristicLab.Common;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.ParallelEngine;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using HeuristicLab.Problems.DataAnalysis.Symbolic.Regression;
using HeuristicLab.Problems.Instances.DataAnalysis;
using HeuristicLab.Random;
using HeuristicLab.Selection;
using HeuristicLab.SequentialEngine;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using ShellProgressBar;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

namespace HeuristicLabConsole
{
  class Program
  {
    const string dataPath = @"E:\Results\Operon\experiment\data";
    const int repetitions = 50;

    static void Main() {
      var dataFiles = Directory.EnumerateFiles(dataPath, "*.json").OrderBy(x => x).ToList();

      foreach (var f in dataFiles) {
        Console.WriteLine(f);
        var json = File.ReadAllText(f);
        var dict = JsonConvert.DeserializeObject<Dictionary<string, object>>(json);
        var meta = (JObject)dict["metadata"];

        var name = meta["name"].Value<string>();
        var fileName = meta["filename"].Value<string>();
        var target = meta["target"].Value<string>();

        var trainingRows = meta["training_rows"].Value<JObject>();
        var trainingStart = trainingRows["start"].Value<int>();
        var trainingEnd = trainingRows["end"].Value<int>();

        var testRows = meta["test_rows"].Value<JObject>();
        var testStart = testRows["start"].Value<int>();
        var testEnd = testRows["end"].Value<int>();

        var provider = new RegressionCSVInstanceProvider();
        var data = provider.ImportData(Path.Combine(dataPath, fileName));
        data.TrainingPartition.Start = trainingStart;
        data.TrainingPartition.End = trainingEnd;
        data.TestPartition.Start = testStart;
        data.TestPartition.End = testEnd;
        data.TargetVariable = target;

        Console.WriteLine($"Training partition: {data.TrainingPartition}");
        Console.WriteLine($"Test partition: {data.TestPartition}");

        var geneticAlgorithm = (GeneticAlgorithm)CreateGA(data);

        Console.WriteLine($"Starting experiment in {Directory.GetCurrentDirectory()}");
        try {
          RunExperiment(geneticAlgorithm, repetitions, "GA_");
        } catch (Exception e) {
          Console.WriteLine($"Failed to run experiment due to exception: {e.Message}");
        }
      }
    }

    static IAlgorithm CreateGA(IRegressionProblemData problemData) {
      var solutionCreator = new SymbolicDataAnalysisExpressionBalancedTreeCreator() { MaximumSymbolicExpressionTreeDepthParameter = { Value = new IntValue(1000) } };
      var problem = new SymbolicRegressionSingleObjectiveProblem {
        ProblemData = (IRegressionProblemData)problemData.Clone(),
        SymbolicExpressionTreeInterpreter = new SymbolicDataAnalysisExpressionTreeNativeInterpreter(),
        MaximumSymbolicExpressionTreeDepth = { Value = 12 },
        MaximumSymbolicExpressionTreeLength = { Value = 50 },
      };
      problem.SolutionCreatorParameter.Value = solutionCreator;
      var grammar = problem.SymbolicExpressionTreeGrammar;
      grammar.Symbols.OfType<Sine>().Single().Enabled = true;
      grammar.Symbols.OfType<Cosine>().Single().Enabled = true;
      grammar.Symbols.OfType<Exponential>().Single().Enabled = true;
      grammar.Symbols.OfType<Logarithm>().Single().Enabled = true;

      var ga = new GeneticAlgorithm {
        Engine = new SequentialEngine(),
        Problem = problem,
        MaximumGenerations = { Value = 1000 },
        PopulationSize = { Value = 1000 },
        MutationProbability = { Value = 0.25 }
      };
      ga.CrossoverParameter.Value = ga.CrossoverParameter.ValidValues.OfType<SubtreeCrossover>().Single();
      var multiMutation = ga.MutatorParameter.ValidValues.OfType<MultiSymbolicExpressionTreeManipulator>().Single();
      foreach (var op in multiMutation.Operators) {
        multiMutation.Operators.SetItemCheckedState(op, op is OnePointShaker || op is ChangeNodeTypeManipulation);
      }
      ga.MutatorParameter.Value = multiMutation;
      var selector = ga.SelectorParameter.ValidValues.OfType<TournamentSelector>().Single();
      selector.GroupSizeParameter.Value.Value = 5;
      ga.SelectorParameter.Value = selector;

      foreach (var op in ga.Analyzer.Operators) {
        ga.Analyzer.Operators.SetItemCheckedState(op, op is SymbolicRegressionSingleObjectiveTrainingBestSolutionAnalyzer);
      }
      return ga;
    }

    static void ExportCsv(IOptimizer optimizer, double elapsed, string path) {
      // aggregate results
      var rows = new List<string> { "Problem", "Generations", "Elapsed" };
      rows.AddRange(new[] { "R2 (train)", "R2 (train) IQR", "R2 (test)", "R2 (test) IQR" });
      rows.AddRange(new[] { "RMSE (train)", "RMSE (train) IQR", "RMSE (test)", "RMSE (test) IQR" });
      rows.AddRange(new[] { "NMSE (train)", "NMSE (train) IQR", "NMSE (test)", "NMSE (test) IQR" });

      string getProblemName(IRun run) => ((IRegressionProblemData)run.Parameters["ProblemData"]).Name;

      using (var sw = new StreamWriter(path)) {
        foreach (var row in rows) {
          sw.Write(row == rows.Last() ? $"{row}\n" : $"{row},");
        }
        var runs = optimizer.Runs;
        var solutions = runs.Select(x => (ISymbolicRegressionSolution)x.Results["Best training solution"]);

        var r2TrainValues = solutions.Select(x => x.TrainingRSquared).ToList();
        var r2Train = r2TrainValues.Median();
        var r2TrainIQR = r2TrainValues.Quantile(0.75) - r2TrainValues.Quantile(0.25);

        var r2TestValues = solutions.Select(x => x.TestRSquared).ToList();
        var r2Test = r2TestValues.Median();
        var r2TestIQR = r2TestValues.Quantile(0.75) - r2TestValues.Quantile(0.25);

        var rmseTrainValues = solutions.Select(x => x.TrainingRootMeanSquaredError).ToList();
        var rmseTrain = rmseTrainValues.Median();
        var rmseTrainIQR = rmseTrainValues.Quantile(0.75) - rmseTrainValues.Quantile(0.25);

        var rmseTestValues = solutions.Select(x => x.TestRootMeanSquaredError).ToList();
        var rmseTest = rmseTestValues.Median();
        var rmseTestIQR = rmseTestValues.Quantile(0.75) - rmseTestValues.Quantile(0.25);

        var nmseTrainValues = solutions.Select(x => x.TrainingNormalizedMeanSquaredError).ToList();
        var nmseTrain = nmseTrainValues.Median();
        var nmseTrainIQR = nmseTrainValues.Quantile(0.75) - nmseTrainValues.Quantile(0.25);

        var nmseTestValues = solutions.Select(x => x.TestNormalizedMeanSquaredError).ToList();
        var nmseTest = nmseTestValues.Median();
        var nmseTestIQR = nmseTestValues.Quantile(0.75) - nmseTestValues.Quantile(0.25);

        var generations = runs.Select(x => (IntValue)x.Results["Generations"]).Select(x => (double)x.Value).Median();

        sw.WriteLine($"{getProblemName(runs.First())},{generations},{elapsed},{r2Train},{r2TrainIQR},{r2Test},{r2TestIQR},{rmseTrain},{rmseTrainIQR},{rmseTest},{rmseTestIQR},{nmseTrain},{nmseTrainIQR},{nmseTest},{nmseTestIQR}");
      }
    }

    static void RunExperiment(IAlgorithm algorithm, int repetitions, string prefix = "") {
      try {
        var proto = new ProtoBufSerializer();
        var problem = (IRegressionProblem)algorithm.Problem;
        var problemData = problem.ProblemData;

        var experiment = new Experiment { NumberOfWorkers = Environment.ProcessorCount };

        int totalTicks = repetitions;
        var stopWatch = new Stopwatch();

        using (var pbar = new ProgressBar(totalTicks, experiment.Name, new ProgressBarOptions { ProgressCharacter = '#', ProgressBarOnBottom = true })) {
          int counter = 0;
          experiment.Optimizers.AddRange(Enumerable.Range(0, repetitions).Select(x => (IAlgorithm)algorithm.Clone()));
          foreach (var optimizer in experiment.Optimizers) {
            optimizer.Stopped += (sender, args) => {
              ++counter;
              pbar.Tick($"Run {counter} out of {experiment.Optimizers.Count}.");

            };
          }

          stopWatch.Start();
          experiment.Start(CancellationToken.None);
          stopWatch.Stop();
        }

        var elapsed = stopWatch.ElapsedMilliseconds / 1000.0;
        Console.WriteLine($"Done in {elapsed} seconds ({elapsed / experiment.Runs.Count} seconds per run). Serializing...");
        proto.Serialize(experiment, prefix + problemData.Name + ".hl");
        ExportCsv(experiment, elapsed / experiment.Runs.Count, prefix + problemData.Name);
      } catch (Exception e) {
        Console.WriteLine(e.Message);
      }
    }
  }
}
